#
# Copyright (c) 2015 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#
from . import resource


class Plan(resource.BaseResource):

    ROOT = 'account/plan'
    PASSTHRU_FIELDS = ['type', 'period', 'notes']

    def retrieve(self, callback=None, errback=None):
        return self._make_request('GET', '%s' % (self.ROOT),
                                  callback=callback,
                                  errback=errback)



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#

from . import resource
try:
    from urllib.parse import urlencode
except:
    from urllib import urlencode


class Stats(resource.BaseResource):

    ROOT = 'stats'

    def qps(self, zone=None, domain=None, type=None,
            callback=None, errback=None):
        url = ''
        if zone is None:
            url = '%s/%s' % (self.ROOT, 'qps')
        elif type is not None and domain is not None and zone is not None:
            url = '%s/%s/%s/%s/%s' % (self.ROOT, 'qps', zone, domain, type)
        elif zone is not None:
            url = '%s/%s/%s' % (self.ROOT, 'qps', zone)
        return self._make_request('GET', url,
                                  callback=callback,
                                  errback=errback)

    def usage(self, zone=None, domain=None, type=None,
              callback=None, errback=None, **kwargs):
        url = ''
        if zone is None:
            url = '%s/%s' % (self.ROOT, 'usage'),
        elif type is not None and domain is not None and zone is not None:
            url = '%s/%s/%s/%s/%s' % (self.ROOT, 'usage', zone, domain, type)
        elif zone is not None:
            url = '%s/%s/%s' % (self.ROOT, 'usage', zone)
        args = {}
        if 'period' in kwargs:
            args['period'] = kwargs['period']
        for f in ['expand', 'aggregate', 'by_tier']:
            if f in kwargs:
                args[f] = bool(kwargs[f])
        return self._make_request('GET', '%s?%s' % (url,
                                                    urlencode(args)),
                                  callback=callback,
                                  errback=errback)



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#

import sys
import logging
import json
from nsone import version
from nsone.rest.transport.base import TransportBase
from nsone.rest.errors import ResourceException


class BaseResource:

    DEFAULT_TRANSPORT = 'requests'

    INT_FIELDS = []
    BOOL_FIELDS = []
    PASSTHRU_FIELDS = []

    def __init__(self, config):
        """

        :param nsone.config.Config config: config object used to build requests
        """
        self._config = config
        self._log = logging.getLogger(__name__)
        # TODO verify we have a default key
        # get a transport. TODO make this static property?
        transport = self._config.get('transport', None)
        if transport is None:
            # for default transport:
            # if requests is available, use that. otherwise, basic
            from nsone.rest.transport.requests import have_requests
            if have_requests:
                transport = 'requests'
            else:
                transport = 'basic'
        if transport not in TransportBase.REGISTRY:
            raise ResourceException('requested transport was not found: %s'
                                    % transport)
        self._transport = TransportBase.REGISTRY[transport](self._config)

    def _buildStdBody(self, body, fields):
        for f in self.BOOL_FIELDS:
            if f in fields:
                body[f] = bool(fields[f])
        for f in self.INT_FIELDS:
            if f in fields:
                body[f] = int(fields[f])
        for f in self.PASSTHRU_FIELDS:
            if f in fields:
                body[f] = fields[f]

    def _make_url(self, path):
        return self._config.getEndpoint() + path

    def _make_request(self, type, path, **kwargs):
        VERBS = ['GET', 'POST', 'DELETE', 'PUT']
        if type not in VERBS:
            raise Exception('invalid request method')
        # TODO don't assume this doesn't exist in kwargs
        kwargs['headers'] = {
            'User-Agent': 'nsone-python %s python 0x%s %s'
                          % (version, sys.hexversion, sys.platform),
            'X-NSONE-Key': self._config.getAPIKey()
        }
        if 'body' in kwargs:
            kwargs['data'] = json.dumps(kwargs['body'])
            del kwargs['body']
        return self._transport.send(type, self._make_url(path), **kwargs)



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#

# flake8: noqa
from nsone.rest.transport.basic import *
# flake8: noqa
from nsone.rest.transport.requests import *
# flake8: noqa
from nsone.rest.transport.twisted import *



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#
from __future__ import absolute_import

from nsone.rest.transport.base import TransportBase
from nsone.rest.errors import ResourceException, RateLimitException, \
    AuthException

try:
    import requests
    have_requests = True
except:
    have_requests = False


class RequestsTransport(TransportBase):

    def __init__(self, config):
        if not have_requests:
            raise ImportError('requests module required for RequestsTransport')
        TransportBase.__init__(self, config, self.__module__)
        self.REQ_MAP = {
            'GET': requests.get,
            'POST': requests.post,
            'DELETE': requests.delete,
            'PUT': requests.put
        }

    def send(self, method, url, headers=None, data=None, files=None,
             callback=None, errback=None):
        self._logHeaders(headers)
        resp = self.REQ_MAP[method](url, headers=headers, verify=self._verify,
                                    data=data, files=files)
        if resp.status_code != 200:
            if errback:
                errback(resp)
                return
            else:
                if resp.status_code == 429:
                    raise RateLimitException('rate limit exceeded',
                                             resp,
                                             resp.text)
                elif resp.status_code == 401:
                    raise AuthException('unauthorized',
                                        resp,
                                        resp.text)
                else:
                    raise ResourceException('server error',
                                            resp,
                                            resp.text)
        # TODO make sure json is valid
        try:
            jsonOut = resp.json()
        except ValueError:
            if errback:
                errback(resp)
                return
            else:
                raise ResourceException('invalid json in response',
                                        resp,
                                        resp.text)
        if callback:
            return callback(jsonOut)
        else:
            return jsonOut

TransportBase.REGISTRY['requests'] = RequestsTransport



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#

from .config import Config

version = "0.9.7"


class NSONE:

    def __init__(self, apiKey=None, config=None, configFile=None, keyID=None):
        """
        Create a new top level NSONE API object

        :param str apiKey: if given, initialize config with this API key \
            (obtainable via creation in NSONE portal)
        :param nsone.config.Config config: if given, uses a separately \
            constructed and configured Config object
        :param str configFile: if given, load configuration from the given \
            json configuration file
        :param str keyID: if given, use the specified key config in the \
            multi-key configuration file
        """
        self.config = config
        if self.config is None:
            self._loadConfig(apiKey, configFile)
        if keyID:
            self.config.useKeyID(keyID)

    def _loadConfig(self, apiKey, configFile):
        self.config = Config()
        if apiKey:
            self.config.createFromAPIKey(apiKey)
        else:
            configFile = Config.DEFAULT_CONFIG_FILE \
                if not configFile else configFile
            self.config.loadFromFile(configFile)

    # REST INTERFACE

    def zones(self):
        """
        Return a new raw REST interface to zone resources

        :rtype: :py:class:`nsone.rest.zones.Zones`
        """
        import nsone.rest.zones
        return nsone.rest.zones.Zones(self.config)

    def records(self):
        """
        Return a new raw REST interface to record resources

        :rtype: :py:class:`nsone.rest.records.Records`
        """
        import nsone.rest.records
        return nsone.rest.records.Records(self.config)

    def stats(self):
        """
        Return a new raw REST interface to stats resources

        :rtype: :py:class:`nsone.rest.stats.Stats`
        """
        import nsone.rest.stats
        return nsone.rest.stats.Stats(self.config)

    def datasource(self):
        """
        Return a new raw REST interface to datasource resources

        :rtype: :py:class:`nsone.rest.data.Source`
        """
        import nsone.rest.data
        return nsone.rest.data.Source(self.config)

    def datafeed(self):
        """
        Return a new raw REST interface to feed resources

        :rtype: :py:class:`nsone.rest.data.Feed`
        """
        import nsone.rest.data
        return nsone.rest.data.Feed(self.config)

    def monitors(self):
        """
        Return a new raw REST interface to monitors resources

        :rtype: :py:class:`nsone.rest.monitoring.Monitors`
        """
        import nsone.rest.monitoring
        return nsone.rest.monitoring.Monitors(self.config)

    def plan(self):
        """
        Return a new raw REST interface to account plan

        :rtype: :py:class:`nsone.rest.account.Plan`
        """
        import nsone.rest.account
        return nsone.rest.account.Plan(self.config)

    # HIGH LEVEL INTERFACE
    def loadZone(self, zone, callback=None, errback=None):
        """
        Load an existing zone into a high level Zone object.

        :param str zone: zone name, like 'example.com'
        :rtype: :py:class:`nsone.zones.Zone`
        """
        import nsone.zones
        zone = nsone.zones.Zone(self.config, zone)
        return zone.load(callback=callback, errback=errback)

    def createZone(self, zone, zoneFile=None, callback=None, errback=None,
                   **kwargs):
        """
        Create a new zone, and return an associated high level Zone object.
        Several optional keyword arguments are available to configure the SOA
        record.

        If zoneFile is specified, upload the specific zone definition file
        to populate the zone with.

        :param str zone: zone name, like 'example.com'
        :param str zoneFile: absolute path of a zone file
        :keyword int retry: retry time
        :keyword int refresh: refresh ttl
        :keyword int expiry: expiry ttl
        :keyword int nx_ttl: nxdomain TTL

        :rtype: :py:class:`nsone.zones.Zone`
        """
        import nsone.zones
        zone = nsone.zones.Zone(self.config, zone)
        return zone.create(zoneFile=zoneFile, callback=callback,
                           errback=errback, **kwargs)

    def loadRecord(self, domain, type, zone=None, callback=None,
                   errback=None, **kwargs):
        """
        Load an existing record into a high level Record object.

        :param str domain: domain name of the record in the zone, for example \
            'myrecord'. You may leave off the zone, since it must be \
            specified in the zone parameter
        :param str type: record type, such as 'A', 'MX', 'AAAA', etc.
        :param str zone: zone name, like 'example.com'
        :rtype: :py:class:`nsone.records`
        """
        import nsone.zones
        if zone is None:
            # extract from record string
            zone = '.'.join(domain.split('.')[1:])
        z = nsone.zones.Zone(self.config, zone)
        return z.loadRecord(domain, type, callback=callback, errback=errback,
                            **kwargs)



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#

from nsone.rest.records import Records
from nsone.rest.stats import Stats


class RecordException(Exception):
    pass


class Record(object):

    """
    High level object representing a Record
    """

    def __init__(self, parentZone, domain, type):
        """
        Create a new high level Record

        :param nsone.zones.Zone parentZone: the high level Zone parent object
        :param str domain: full domain name this record represents. if the \
          domain does not end with the zone name, it is appended.
        :param str type: The DNS record type (A, MX, etc)
        """
        self._rest = Records(parentZone.config)
        self.parentZone = parentZone
        if not domain.endswith(parentZone.zone):
            domain = domain + '.' + parentZone.zone
        self.domain = domain
        self.type = type
        self.data = None

    def __repr__(self):
        return '<Record domain=%s type=%s>' % (self.domain, self.type)

    def __getitem__(self, item):
        return self.data.get(item, None)

    def _parseModel(self, data):
        self.data = data
        self.answers = data['answers']
        # XXX break out the rest? use getattr instead?

    def reload(self, callback=None, errback=None):
        """
        Reload record data from the API.
        """
        return self.load(reload=True, callback=callback, errback=errback)

    def load(self, callback=None, errback=None, reload=False):
        """
        Load record data from the API.
        """
        if not reload and self.data:
            raise RecordException('record already loaded')

        def success(result, *args):
            self._parseModel(result)
            if callback:
                return callback(self)
            else:
                return self
        return self._rest.retrieve(self.parentZone.zone,
                                   self.domain, self.type,
                                   callback=success, errback=errback)

    def delete(self, callback=None, errback=None):
        """
        Delete the record from the zone, including all advanced configuration,
        meta data, etc.
        """
        if not self.data:
            raise RecordException('record not loaded')

        def success(result, *args):
            if callback:
                return callback(result)
            else:
                return result

        return self._rest.delete(self.parentZone.zone,
                                 self.domain, self.type,
                                 callback=success, errback=errback)

    def update(self, callback=None, errback=None, **kwargs):
        """
        Update record configuration. Pass list of keywords and their values to
        update. For the list of keywords available for zone configuration, see
        :attr:`nsone.rest.records.Records.INT_FIELDS`,
        :attr:`nsone.rest.records.Records.PASSTHRU_FIELDS`,
        :attr:`nsone.rest.records.Records.BOOL_FIELDS`
        """
        if not self.data:
            raise RecordException('record not loaded')

        def success(result, *args):
            self._parseModel(result)
            if callback:
                return callback(self)
            else:
                return self

        return self._rest.update(self.parentZone.zone, self.domain, self.type,
                                 callback=success, errback=errback, **kwargs)

    def create(self, callback=None, errback=None, **kwargs):
        """
        Create new record. Pass a list of keywords and their values to
        config. For the list of keywords available for zone configuration, see
        :attr:`nsone.rest.records.Records.INT_FIELDS`,
        :attr:`nsone.rest.records.Records.PASSTHRU_FIELDS`,
        :attr:`nsone.rest.records.Records.BOOL_FIELDS`
        """
        if self.data:
            raise RecordException('record already loaded')

        def success(result, *args):
            self._parseModel(result)
            if callback:
                return callback(self)
            else:
                return self

        return self._rest.create(self.parentZone.zone, self.domain, self.type,
                                 callback=success, errback=errback, **kwargs)

    def qps(self, callback=None, errback=None):
        """
        Return the current QPS for this record

        :rtype: dict
        :return: QPS information
        """
        if not self.data:
            raise RecordException('record not loaded')
        stats = Stats(self.parentZone.config)
        return stats.qps(zone=self.parentZone.zone,
                         domain=self.domain,
                         type=self.type,
                         callback=callback,
                         errback=errback)

    def usage(self, callback=None, errback=None, **kwargs):
        """
        Return the current usage information for this record

        :rtype: dict
        :return: usage information
        """
        if not self.data:
            raise RecordException('record not loaded')
        stats = Stats(self.parentZone.config)
        return stats.usage(zone=self.parentZone.zone,
                           domain=self.domain,
                           type=self.type,
                           callback=callback,
                           errback=errback,
                           **kwargs)

    def addAnswers(self, answers, callback=None, errback=None, **kwargs):
        """
        Add answers to the record.

        :param answers: answers structure. See the class note on answer format.
        """
        if not self.data:
            raise RecordException('record not loaded')
        orig_answers = self.data['answers']
        new_answers = self._rest._getAnswersForBody(answers)
        orig_answers.extend(new_answers)
        return self.update(answers=orig_answers, callback=callback,
                           errback=errback, **kwargs)



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#
import copy
import logging


class TransportBase(object):

    REGISTRY = {}

    def __init__(self, config, module):
        self._config = config
        self._log = logging.getLogger(module)
        self._verify = not self._config.getKeyConfig() \
            .get('ignore-ssl-errors',
                 self._config.get(
                     'ignore-ssl-errors',
                     False))

    def _logHeaders(self, headers):
        if self._config['verbosity'] > 0:
            argcopy = copy.deepcopy(headers)
            argcopy['X-NSONE-Key'] = '<redacted>'
            self._log.debug(argcopy)

    def send(self, method, url, headers=None, data=None, files=None,
             callback=None, errback=None):
        pass



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#
from . import resource


class Monitors(resource.BaseResource):

    ROOT = 'monitoring/jobs'
    PASSTHRU_FIELDS = ['name', 'config']

    def list(self, callback=None, errback=None):
        return self._make_request('GET', '%s' % (self.ROOT),
                                  callback=callback,
                                  errback=errback)

    def update(self, jobid, body, callback=None, errback=None, **kwargs):
        self._buildStdBody(body, kwargs)
        return self._make_request('POST',
                                  '%s/%s' % (self.ROOT, jobid),
                                  body=body,
                                  callback=callback,
                                  errback=errback)

    def create(self,body, callback=None, errback=None):
        return self._make_request('PUT', '%s' % (self.ROOT), body=body,
                                  callback=callback,
                                  errback=errback)

    def retrieve(self, jobid, callback=None, errback=None):
        return self._make_request('GET', '%s/%s' % (self.ROOT, jobid),
                                  callback=callback,
                                  errback=errback)

    def delete(self, jobid, callback=None, errback=None):
        return self._make_request('DELETE', '%s/%s' % (self.ROOT, jobid),
                                  callback=callback,
                                  errback=errback)



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
from __future__ import absolute_import

from nsone.rest.transport.base import TransportBase
from nsone.rest.errors import ResourceException, RateLimitException, \
    AuthException
import json
import random

try:
    import StringIO
except ImportError:  # Python 3 +
    from io import StringIO

try:
    from twisted.internet import reactor
    from twisted.web.client import Agent, readBody, FileBodyProducer
    from twisted.web.http_headers import Headers
    from twisted.internet.defer import succeed
    have_twisted = True
except Exception as e:
    have_twisted = False


def encodeForm(varname, f, ctype):
    randomChars = [str(random.randrange(10)) for _ in range(28)]
    boundary = "".join(randomChars)

    lines = ['--' + boundary]

    def add(name, content):
        header = 'Content-Disposition: form-data; name="%s"; filename="%s"' % \
                 (name, f.name)
        header += ";\r\nContent-Type: %s" % ctype
        lines.extend([header, "", content])

    add(varname, f.read())

    lines.extend(['--' + boundary + "--", ""])
    return boundary, "\r\n".join(lines)


class StringProducer(object):

    def __init__(self, body):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)
        return succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


class TwistedTransport(TransportBase):

    def __init__(self, config):
        if not have_twisted:
            raise ImportError('Twisted required for TwistedTransport')
        TransportBase.__init__(self, config, self.__module__)
        self.agent = Agent(reactor)

    def _callback(self, response, user_callback, data, headers):
        d = readBody(response)
        d.addCallback(self._onBody, response, user_callback, data, headers)
        return d

    def _onBody(self, body, response, user_callback, data, headers):
        self._logHeaders(headers)
        self._log.debug("%s %s %s %s" % (response.request.method,
                                         response.request.absoluteURI,
                                         response.code,
                                         data))
        if response.code != 200:
            if response.code == 429:
                raise RateLimitException('rate limit exceeded',
                                         response,
                                         body)
            elif response.code == 401:
                raise AuthException('unauthorized',
                                    response,
                                    body)
            else:
                raise ResourceException('server error', response, body)
        try:
            jsonOut = json.loads(body)
        except:
            raise ResourceException('invalid json in response',
                                    response,
                                    body)
        if user_callback:
            # set these in case callback throws, so we have them for errback
            self.response = response
            self.body = body
            return user_callback(jsonOut, body, response)
        else:
            return jsonOut

    def _errback(self, failure, user_errback):
        # print "failure: %s" % failure.printTraceback()
        if user_errback:
            return user_errback(failure)
        else:
            if failure.check(ResourceException, RateLimitException):
                raise failure.value
            raise ResourceException(failure.getErrorMessage())

    def send(self, method, url, headers=None, data=None, files=None,
             callback=None, errback=None):
        bProducer = None
        if data:
            bProducer = StringProducer(data)
        elif files:
            if len(files) > 1:
                raise Exception('twisted transport currently only accepts one'
                                ' multipart file')
            boundary, body = encodeForm(files[0][0],
                                        files[0][1][1],
                                        files[0][1][2])
            if headers is None:
                headers = {}
            headers['Content-Type'] =\
                "multipart/form-data; boundary={}".format(boundary)
            bProducer = FileBodyProducer(StringIO.StringIO(body))
        theaders = None
        if headers:
            theaders = Headers({str(k): [str(v)]
                                for (k, v) in headers.iteritems()})
        d = self.agent.request(method, str(url), theaders, bProducer)
        d.addCallback(self._callback, callback, data, headers)
        d.addErrback(self._errback, errback)
        return d

TransportBase.REGISTRY['twisted'] = TwistedTransport



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#
from __future__ import absolute_import

from nsone.rest.transport.base import TransportBase
from nsone.rest.errors import ResourceException, RateLimitException, \
    AuthException

try:
    from urllib.request import build_opener, Request, HTTPSHandler
    from urllib.error import HTTPError
except:
    from urllib2 import build_opener, Request, HTTPSHandler
    from urllib2 import HTTPError
import json


class BasicTransport(TransportBase):

    def __init__(self, config):
        TransportBase.__init__(self, config, self.__module__)

    def send(self, method, url, headers=None, data=None, files=None,
             callback=None, errback=None):
        if files is not None:
            # XXX
            raise Exception('file uploads not supported in BasicTransport yet')
        self._logHeaders(headers)
        self._log.debug("%s %s %s" % (method, url, data))
        opener = build_opener(HTTPSHandler)
        request = Request(url, headers=headers, data=data)
        request.get_method = lambda: method

        def handleProblem(code, resp, msg):
            if errback:
                errback((resp, msg))
                return

            if code == 429:
                raise RateLimitException('rate limit exceeded',
                                         resp,
                                         msg)
            elif code == 401:
                raise AuthException('unauthorized',
                                    resp,
                                    msg)
            else:
                raise ResourceException('server error',
                                        resp,
                                        msg)

        # Handle error and responses the same so we can
        # always pass the body to the handleProblem function
        try:
            resp = opener.open(request)
        except HTTPError as e:
            resp = e
        finally:
            body = resp.read()
            if resp.code != 200:
                handleProblem(resp.code, resp, body)

        # TODO make sure json is valid
        try:
            jsonOut = json.loads(body)
        except ValueError:
            if errback:
                errback(resp)
                return
            else:
                raise ResourceException('invalid json in response',
                                        resp,
                                        resp.text)
        if callback:
            return callback(jsonOut)
        else:
            return jsonOut

TransportBase.REGISTRY['basic'] = BasicTransport



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#

import json


class ResourceException(Exception):

    def __init__(self, message, response=None, body=None):
        # if message is json error message, unwrap the actual message
        # otherwise, fall back to the whole body
        if body:
            try:
                jData = json.loads(body)
                self.message = '%s: %s' % (message, jData['message'])
            except:
                self.message = message
        else:
            self.message = message
        self.response = response
        self.body = body

    def __repr__(self):
        m = self.message or 'empty message'
        r = self.response or 'empty response'
        if self.body and len(self.body) > 30:
            b = '%s...' % self.body[0:30]
        else:
            b = self.body or 'empty body'
        return '<ResourceException message=%s, response=%s, body=%s>' % \
               (m, r, b)

    def __str__(self):
        return self.message


class AuthException(ResourceException):

    def __repr__(self):
        return '<AuthException>'

    def __str__(self):
        return 'unauthorized'


class RateLimitException(ResourceException):

    def __init__(self, message, response=None, body=None):
        ResourceException.__init__(self, message, response, body)
        hdrs = response.headers._rawHeaders
        self.by = hdrs['x-ratelimit-by'][0]
        self.limit = hdrs['x-ratelimit-limit'][0]
        self.period = hdrs['x-ratelimit-period'][0]

    def __repr__(self):
        return '<RateLimitException by=%s limit=%s period=%s>' % \
               (self.by, self.limit, self.period)



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#

from nsone.rest.zones import Zones
from nsone.records import Record
from nsone.rest.stats import Stats
from nsone.rest.records import Records


class ZoneException(Exception):
    pass


class Zone(object):

    """
    High level object representing a Zone. In addition to the documented
    methods, there are magic methods allowing easy creation of records in this
    zone. Simply can 'add_TYPE' where TYPE is a valid DNS record type, such as
    add_A(). See examples for more information.
    """

    def __init__(self, config, zone):
        """
        Create a new high level Zone object

        :param nsone.config.Config config: config object
        :param str zone: zone name
        """
        self._rest = Zones(config)
        self.config = config
        self.zone = zone
        self.data = None

    def __repr__(self):
        return '<Zone zone=%s>' % self.zone

    def __getitem__(self, item):
        return self.data.get(item, None)

    def reload(self, callback=None, errback=None):
        """
        Reload zone data from the API.
        """
        return self.load(reload=True, callback=callback, errback=errback)

    def load(self, callback=None, errback=None, reload=False):
        """
        Load zone data from the API.
        """
        if not reload and self.data:
            raise ZoneException('zone already loaded')

        def success(result, *args):
            self.data = result
            if callback:
                return callback(self)
            else:
                return self

        return self._rest.retrieve(self.zone, callback=success,
                                   errback=errback)

    def delete(self, callback=None, errback=None):
        """
        Delete the zone and ALL records it contains.
        """
        return self._rest.delete(self.zone, callback=callback, errback=errback)

    def update(self, callback=None, errback=None, **kwargs):
        """
        Update zone configuration. Pass a list of keywords and their values to
        update. For the list of keywords available for zone configuration, see
        :attr:`nsone.rest.zones.Zones.INT_FIELDS` and
        :attr:`nsone.rest.zones.Zones.PASSTHRU_FIELDS`
        """
        if not self.data:
            raise ZoneException('zone not loaded')

        def success(result, *args):
            self.data = result
            if callback:
                return callback(self)
            else:
                return self

        return self._rest.update(self.zone, callback=success, errback=errback,
                                 **kwargs)

    def create(self, zoneFile=None, callback=None, errback=None, **kwargs):
        """
        Create a new zone. Pass a list of keywords and their values to
        configure. For the list of keywords available for zone configuration,
        see :attr:`nsone.rest.zones.Zones.INT_FIELDS` and
        :attr:`nsone.rest.zones.Zones.PASSTHRU_FIELDS`
        If zoneFile is passed, it should be a zone text file on the local disk
        that will be used to populate the created zone file.
        """
        if self.data:
            raise ZoneException('zone already loaded')

        def success(result, *args):
            self.data = result
            if callback:
                return callback(self)
            else:
                return self

        if zoneFile:
            return self._rest.import_file(self.zone, zoneFile,
                                          callback=success, errback=errback,
                                          **kwargs)
        else:
            return self._rest.create(self.zone, callback=success,
                                     errback=errback, **kwargs)

    def __getattr__(self, item):

        if not item.startswith('add_'):
            raise AttributeError(item)

        # dynamic adding of various record types, e.g. add_A, add_CNAME, etc
        (_, rtype) = item.split('_', 2)

        def add_X(domain, answers, callback=None, errback=None, **kwargs):
            kwargs['answers'] = answers
            record = Record(self, domain, rtype)
            return record.create(callback=callback, errback=errback, **kwargs)
        return add_X

    def createLinkToSelf(self, new_zone, callback=None, errback=None,
                         **kwargs):
        """
        Create a new linked zone, linking to ourselves. All records in this
        zone will then be available as "linked records" in the new zone.

        :param str new_zone: the new zone name to link to this one
        :return: new Zone
        """
        zone = Zone(self.config, new_zone)
        kwargs['link'] = self.data['zone']
        return zone.create(callback=callback, errback=errback, **kwargs)

    def linkRecord(self, existing_domain, new_domain, rtype,
                   callback=None, errback=None, **kwargs):

        """
        Create a new linked record in this zone. These records use the
        configuration (answers, ttl, filters, etc) from an existing record
        in the NSONE platform.

        :param str existing_domain: FQDN of the target record whose config \
            should be used. Does not have to be in the same zone.
        :param str new_domain: Name of the new (linked) record. Zone name is\
            appended automatically.
        :param str rtype: DNS record type, which must match the target record.
        :rtype: nsone.records.Record
        :return: new Record
        """

        if '.' not in existing_domain:
            existing_domain = existing_domain + '.' + self.zone

        record = Record(self, new_domain, rtype)
        return record.create(answers=[],
                             link=existing_domain,
                             callback=callback,
                             errback=errback,
                             **kwargs)

    def cloneRecord(self, existing_domain, new_domain, rtype,
                    zone=None, callback=None, errback=None):
        """
        Clone the given record to a new record such that their configs are
        identical.

        :param str existing_domain: The existing record to clone
        :param str new_domain: The name of the new cloned record
        :param str rtype: DNS record type
        :param str zone: Optional zone name, if the new record should exist in\
            a different zone than the original record.
        :rtype: nsone.records.Record
        :return: new Record
        """
        if zone is None:
            zone = self.zone

        if not new_domain.endswith(zone):
            new_domain = new_domain + '.' + zone

        def onSaveNewRecord(new_data):
            if zone != self.zone:
                pZone = Zone(self.config, zone)
            else:
                pZone = self
            new_rec = Record(pZone, new_domain, rtype)
            new_rec._parseModel(new_data)
            if callback:
                return callback(new_rec)
            else:
                return new_rec

        def onLoadRecord(old_rec):
            data = old_rec.data
            data['zone'] = zone
            data['domain'] = new_domain
            restapi = Records(self.config)
            return restapi.create_raw(zone, new_domain, rtype, data,
                                      callback=onSaveNewRecord,
                                      errback=errback)

        return self.loadRecord(existing_domain, rtype,
                               callback=onLoadRecord,
                               errback=errback)

    def loadRecord(self, domain, rtype, callback=None, errback=None):
        """
        Load a high level Record object from a domain within this Zone.

        :param str domain: The name of the record to load
        :param str rtype: The DNS record type
        :rtype: nsone.records.Record
        :return: new Record
        """
        rec = Record(self, domain, rtype)
        return rec.load(callback=callback, errback=errback)

    def qps(self, callback=None, errback=None):
        """
        Return the current QPS for this zone

        :rtype: dict
        :return: QPS information
        """
        stats = Stats(self.config)
        return stats.qps(zone=self.zone, callback=callback, errback=errback)

    def usage(self, callback=None, errback=None, **kwargs):
        """
        Return the current usage information for this zone

        :rtype: dict
        :return: usage information
        """
        stats = Stats(self.config)
        return stats.usage(zone=self.zone, callback=callback, errback=errback,
                           **kwargs)



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#

import json
import os


class ConfigException(Exception):
    pass


class Config:

    """A simple object for accessing and manipulating config files. These
    contains options and credentials for accessing the NSONE REST API.
    Config files are simple JSON text files.
    To set or retrieve vales, use the object like a dict.
    """
    ENDPOINT = 'api.nsone.net'

    PORT = 443

    API_VERSION = 'v1'

    DEFAULT_CONFIG_FILE = '~/.nsone'

    def __init__(self, path=None):
        """
        :param str path: optional path. if given, try to load the given config\
            file
        """
        self._path = None
        self._keyID = None
        self._data = {}
        if path:
            self.loadFromFile(path)

    def _doDefaults(self):
        if 'default_key' in self._data:
            self.useKeyID(self._data['default_key'])
        if 'endpoint' not in self._data:
            self._data['endpoint'] = self.ENDPOINT
        if 'port' not in self._data:
            self._data['port'] = self.PORT
        if 'api_version' not in self._data:
            self._data['api_version'] = self.API_VERSION
        if 'cli' not in self._data:
            self._data['cli'] = {}
        if 'verbosity' not in self._data:
            self._data['verbosity'] = 0

    def createFromAPIKey(self, apikey, maybeWriteDefault=False):
        """
        Create a basic config from a single API key

        :param str apikey: NSONE API Key, as created in the NSONE portal
        :param bool maybeWriteDefault: If True and DEFAULT_CONFIG_FILE doesn't\
            exist write out the resulting config there.
        """
        self._data = {
            'default_key': 'default',
            'keys': {
                'default': {
                    'key': apikey,
                    'desc': 'imported API key'
                }
            }
        }
        self._keyID = 'default'
        self._doDefaults()
        if maybeWriteDefault:
            path = os.path.expanduser(self.DEFAULT_CONFIG_FILE)
            self.write(path)

    def loadFromDict(self, d):
        """
        Load config data from the given dictionary

        :param dict d: Python dictionary containing configuration items
        """
        self._data = d
        self._doDefaults()

    def loadFromString(self, body):
        """
        Load config data (i.e. JSON text) from the given string

        :param str body: config data in JSON format
        """
        try:
            self._data = json.loads(body)
        except Exception as e:
            raise ConfigException('%s: invalid config body: %s' %
                                  (self._path, e.message))
        self._doDefaults()

    def loadFromFile(self, path):
        """
        Load JSON config file from disk at the given path

        :param str path: path to config file
        """
        if '~' in path:
            path = os.path.expanduser(path)
        f = open(path)
        body = f.read()
        f.close()
        self._path = path
        self.loadFromString(body)

    def write(self, path=None):
        """
         Write config data to disk. If this config object already has a path,
         it will write to it. If it doesn't, one must be passed during this
         call.

        :param str path: path to config file
        """
        if not self._path and not path:
            raise ConfigException('no config path given')
        if path:
            self._path = path
        if '~' in self._path:
            self._path = os.path.expanduser(self._path)
        f = open(self._path, 'w')
        f.write(json.dumps(self._data))
        f.close()

    def useKeyID(self, keyID):
        """
        Use the given API key config specified by `keyID` during subsequent
        API calls

        :param str keyID: an index into the 'keys' maintained in this config
        """
        if keyID not in self._data['keys']:
            raise ConfigException('keyID does not exist: %s' % keyID)
        self._keyID = keyID

    def getCurrentKeyID(self):
        """
        Retrieve the current keyID in use.

        :return: current keyID in use
        """
        return self._keyID

    def getKeyConfig(self, keyID=None):
        """
        Get key configuration specified by `keyID`, or current keyID.

        :param str keyID: optional keyID to retrieve, or current if not passed
        :return: a dict of the request (or current) key config
        """
        k = keyID if keyID is not None else self._keyID
        if not k or k not in self._data['keys']:
            raise ConfigException('request key does not exist: %s' % k)
        return self._data['keys'][k]

    def isKeyWriteLocked(self, keyID=None):
        """
        Determine if a key config is write locked.

        :param str keyID: optional keyID to retrieve, or current if not passed
        :return: True if the given (or current) keyID is writeLocked
        """
        kcfg = self.getKeyConfig(keyID)
        return 'writeLock' in kcfg and kcfg['writeLock'] is True

    def getAPIKey(self, keyID=None):
        """
        Retrieve the NSONE API Key for the given keyID

        :param str keyID: optional keyID to retrieve, or current if not passed
        :return: API Key for the given keyID
        """
        kcfg = self.getKeyConfig(keyID)
        if 'key' not in kcfg:
            raise ConfigException('invalid config: missing api key')
        return kcfg['key']

    def getEndpoint(self):
        """
        Retrieve the NSONE API Endpoint URL that will be used for requests.

        :return: URL of the NSONE API that will be used for requests
        """
        port = ''
        endpoint = ''
        keyConfig = self.getKeyConfig()
        if 'port' in keyConfig:
            port = ':' + keyConfig['port']
        elif self._data['port'] != self.PORT:
            port = ':' + self._data['port']
        if 'endpoint' in keyConfig:
            endpoint = keyConfig['endpoint']
        else:
            endpoint = self._data['endpoint']
        return 'https://%s%s/%s/' % (endpoint,
                                     port,
                                     self._data['api_version'])

    def __repr__(self):
        return 'config file [%s]: %s' % (self._path,
                                         json.dumps(self._data, indent=True))

    def __getitem__(self, item):
        return self._data.get(item, None)

    def __setitem__(self, key, value):
        self._data[key] = value

    def get(self, item, default=None):
        """
        Retrieve a value from the config object.

        :param str item: Key to lookup
        :param default: Default value to return if the requested item doesn't \
            exist
        :return: Requested value, or `default` if it didn't exist
        """
        return self._data.get(item, default)



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#

from . import resource


class Zones(resource.BaseResource):

    ROOT = 'zones'

    INT_FIELDS = ['retry', 'refresh', 'expiry', 'nx_ttl']
    PASSTHRU_FIELDS = ['secondary', 'hostmaster', 'meta', 'networks', 'link']

    def _buildBody(self, zone, **kwargs):
        body = {}
        body['zone'] = zone
        self._buildStdBody(body, kwargs)
        return body

    def import_file(self, zone, zoneFile, callback=None, errback=None,
                    **kwargs):
        files = [('zonefile', (zoneFile, open(zoneFile, 'rb'), 'text/plain'))]
        return self._make_request('PUT',
                                  'import/zonefile/%s' % (zone),
                                  files=files,
                                  callback=callback,
                                  errback=errback)

    def create(self, zone, callback=None, errback=None, **kwargs):
        body = self._buildBody(zone, **kwargs)
        return self._make_request('PUT',
                                  '%s/%s' % (self.ROOT, zone),
                                  body=body,
                                  callback=callback,
                                  errback=errback)

    def update(self, zone, callback=None, errback=None, **kwargs):
        body = self._buildBody(zone, **kwargs)
        return self._make_request('POST',
                                  '%s/%s' % (self.ROOT, zone),
                                  body=body,
                                  callback=callback,
                                  errback=errback)

    def delete(self, zone, callback=None, errback=None):
        return self._make_request('DELETE', '%s/%s' % (self.ROOT, zone),
                                  callback=callback,
                                  errback=errback)

    def list(self, callback=None, errback=None):
        return self._make_request('GET', '%s' % self.ROOT,
                                  callback=callback,
                                  errback=errback)

    def retrieve(self, zone, callback=None, errback=None):
        return self._make_request('GET', '%s/%s' % (self.ROOT, zone),
                                  callback=callback,
                                  errback=errback)



import os
import sys
import nsone

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

path, script = os.path.split(sys.argv[0])
os.chdir(os.path.abspath(path))

tests_require = [
    'pytest',
    'pytest-pep8',
    'pytest-cov',
    'mock',
]

setup(
    name='nsone',
    # flake8: noqa
    version=nsone.version,
    description='Python SDK for the NSONE DNS platform',
    author='Shannon Weyrick',
    author_email='sweyrick@nsone.net',
    url='https://github.com/nsone/nsone-python',
    packages=['nsone', 'nsone.rest', 'nsone.rest.transport'],
    setup_requires=['pytest-runner'],
    tests_require=tests_require,
    keywords='dns development rest sdk nsone',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet :: Name Service (DNS)"
    ])



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.
#
from . import resource


class Source(resource.BaseResource):

    ROOT = 'data/sources'
    PASSTHRU_FIELDS = ['name', 'config']

    def list(self, callback=None, errback=None):
        return self._make_request('GET', '%s' % (self.ROOT),
                                  callback=callback,
                                  errback=errback)

    def retrieve(self, sourceid, callback=None, errback=None):
        return self._make_request('GET',
                                  '%s/%s' % (self.ROOT, sourceid),
                                  callback=callback,
                                  errback=errback)

    def create(self, name, sourcetype, callback=None, errback=None, **kwargs):
        body = {
            'name': name,
            'sourcetype': sourcetype
        }
        self._buildStdBody(body, kwargs)
        return self._make_request('PUT',
                                  '%s' % (self.ROOT),
                                  body=body,
                                  callback=callback,
                                  errback=errback)

    def update(self, sourceid, callback=None, errback=None, **kwargs):
        body = {
            'id': sourceid
        }
        self._buildStdBody(body, kwargs)
        return self._make_request('POST',
                                  '%s/%s' % (self.ROOT, sourceid),
                                  body=body,
                                  callback=callback,
                                  errback=errback)

    def delete(self, sourceid, callback=None, errback=None):
        return self._make_request('DELETE',
                                  '%s/%s' % (self.ROOT, sourceid),
                                  callback=callback,
                                  errback=errback)

    def publish(self, sourceid, data, callback=None, errback=None):
        return self._make_request('POST',
                                  'feed/%s' % (sourceid),
                                  body=data,
                                  callback=callback,
                                  errback=errback)


class Feed(resource.BaseResource):

    ROOT = 'data/feeds'
    PASSTHRU_FIELDS = ['name', 'config']

    def list(self, sourceid, callback=None, errback=None):
        return self._make_request('GET',
                                  '%s/%s' % (self.ROOT, sourceid),
                                  callback=callback,
                                  errback=errback)

    def retrieve(self, sourceid, feedid, callback=None, errback=None):
        return self._make_request('GET',
                                  '%s/%s/%s' % (self.ROOT, sourceid, feedid),
                                  callback=callback,
                                  errback=errback)

    def create(self, sourceid, name, config, callback=None, errback=None,
               **kwargs):
        body = {
            'name': name,
            'config': config,
        }
        self._buildStdBody(body, kwargs)
        return self._make_request('PUT',
                                  '%s/%s' % (self.ROOT, sourceid),
                                  body=body,
                                  callback=callback,
                                  errback=errback)

    def update(self, sourceid, feedid, callback=None, errback=None, **kwargs):
        body = {
            'id': feedid
        }
        self._buildStdBody(body, kwargs)
        return self._make_request('POST',
                                  '%s/%s/%s' % (self.ROOT, sourceid, feedid),
                                  body=body,
                                  callback=callback,
                                  errback=errback)

    def delete(self, sourceid, feedid, callback=None, errback=None):
        return self._make_request('DELETE',
                                  '%s/%s/%s' % (self.ROOT, sourceid, feedid),
                                  callback=callback,
                                  errback=errback)



#
# Copyright (c) 2014 NSONE, Inc.
#
# License under The MIT License (MIT). See LICENSE in project root.

import sys
import collections

from . import resource


py_str = str if sys.version_info[0] == 3 else basestring


class Records(resource.BaseResource):

    ROOT = 'zones'

    INT_FIELDS = ['ttl']
    BOOL_FIELDS = ['use_csubnet']
    PASSTHRU_FIELDS = ['networks', 'meta', 'regions', 'link']

    # answers must be:
    #  1) a single string
    #     we coerce to a single answer with no other fields e.g. meta
    #  2) an iterable of single strings
    #     we coerce to several answers with no other fields e.g. meta
    #  3) an iterable of iterables
    #     we have as many answers as are in the outer iterable, and the
    #     answers themselves are used verbatim from the inner iterable (e.g. may
    #     have MX style [10, '1.1.1.1']), but no other fields e.g. meta
    #     you must use this form for MX records, and if there is only one
    #     answer it still must be wrapped in an outer iterable
    #  4) an iterable of dicts
    #     we assume the full rest model and pass it in unchanged. must use this
    #     form for any advanced record config like meta data or data feeds
    def _getAnswersForBody(self, answers):
        realAnswers = []
        # simplest: they specify a single string ip
        if isinstance(answers, py_str):
            answers = [answers]
        # otherwise, we need an iterable
        elif not isinstance(answers, collections.Iterable):
            raise Exception('invalid answers format (must be str or iterable)')
        # at this point we have a list. loop through and build out the answer
        # entries depending on contents
        for a in answers:
            if isinstance(a, py_str):
                realAnswers.append({'answer': [a]})
            elif isinstance(a, (list, tuple)):
                realAnswers.append({'answer': a})
            elif isinstance(a, dict):
                realAnswers.append(a)
            else:
                raise Exception('invalid answers format: list must contain '
                                'only str, list, or dict')
        return realAnswers

    # filters must be a list of dict which can have two forms:
    # 1) simple: each item in list is a dict with a single key and value. the
    #            key is the name of the filter, the value is a dict of config
    #            values (which may be empty {})
    # 2) full: each item in the list is a dict of the full rest model for
    #          filters (documented elsewhere) which is passed through. use this
    #          for enabled/disabled or future fields not supported otherwise
    #
    def _getFiltersForBody(self, filters):
        realFilters = []

        if type(filters) is not list:
            raise Exception('filter argument must be list of dict')

        for f in filters:
            if type(f) is not dict:
                raise Exception('filter items must be dict')

            if 'filter' in f:
                # full
                realFilters.append(f)
            else:
                # simple, synthesize
                (fname, fconfig) = f.popitem()
                realFilters.append({'filter': fname, 'config': fconfig})

        return realFilters

    def _buildBody(self, zone, domain, type, **kwargs):
        body = {}
        body['zone'] = zone
        body['domain'] = domain
        body['type'] = type.upper()

        if 'filters' in kwargs:
            body['filters'] = self._getFiltersForBody(kwargs['filters'])

        if 'answers' in kwargs:
            body['answers'] = self._getAnswersForBody(kwargs['answers'])

        self._buildStdBody(body, kwargs)

        if 'use_csubnet' in body:
            # key mapping
            body['use_client_subnet'] = body['use_csubnet']
            del body['use_csubnet']

        return body

    def create(self, zone, domain, type,
               callback=None, errback=None, **kwargs):
        body = self._buildBody(zone, domain, type, **kwargs)
        return self.create_raw(zone, domain, type, body,
                               callback=callback, errback=errback,
                               **kwargs)

    def create_raw(self, zone, domain, type, body,
                   callback=None, errback=None, **kwargs):
        return self._make_request('PUT',
                                  '%s/%s/%s/%s' % (self.ROOT,
                                                   zone,
                                                   domain,
                                                   type.upper()),
                                  body=body,
                                  callback=callback,
                                  errback=errback)

    def update(self, zone, domain, type,
               callback=None, errback=None, **kwargs):
        body = self._buildBody(zone, domain, type, **kwargs)
        return self._make_request('POST',
                                  '%s/%s/%s/%s' % (self.ROOT,
                                                   zone,
                                                   domain,
                                                   type.upper()),
                                  body=body,
                                  callback=callback,
                                  errback=errback)

    def delete(self, zone, domain, type, callback=None, errback=None):
        return self._make_request('DELETE', '%s/%s/%s/%s' %
                                  (self.ROOT,
                                   zone,
                                   domain,
                                   type.upper()),
                                  callback=callback,
                                  errback=errback)

    def retrieve(self, zone, domain, type, callback=None, errback=None):
        return self._make_request('GET', '%s/%s/%s/%s' %
                                  (self.ROOT,
                                   zone,
                                   domain,
                                   type.upper()),
                                  callback=callback,
                                  errback=errback)
